<?php

namespace Drupal\entity_workflow\WorkflowType;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\entity_workflow\EntityWorkflowState;
use Drupal\entity_workflow\EntityWorkflowTransition;
use Drupal\workflows\Plugin\WorkflowTypeBase;
use Drupal\workspaces\WorkspaceManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines a base class for Entity Workflow type plugins.
 */
abstract class EntityWorkflowTypeBase extends WorkflowTypeBase implements EntityWorkflowTypeInterface, ContainerFactoryPluginInterface {

  use StringTranslationTrait;
  use NonConfigurableWorkflowTypeTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The workspaces manager.
   *
   * @var \Drupal\workspaces\WorkspaceManagerInterface
   */
  protected $workspaceManager;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * Constructs a EntityWorkflowTypeBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\workspaces\WorkspaceManagerInterface $workspace_manager
   *   The workspace manager.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, WorkspaceManagerInterface $workspace_manager, ModuleHandlerInterface $module_handler) {
    $this->entityTypeManager = $entity_type_manager;
    $this->workspaceManager = $workspace_manager;
    $this->moduleHandler = $module_handler;

    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('workspaces.manager'),
      $container->get('module_handler')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'entity_types' => [],
      'default_state' => NULL,
      'default_state_callback' => NULL,
      'default_transition' => NULL,
      'default_transition_log' => NULL,
      'states' => [],
      'transitions' => [],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    parent::setConfiguration($configuration);

    // Allow modules to alter the workflow type configuration
    $this->moduleHandler->alter('entity_workflow_type', $this->configuration, $this->pluginId);

    if ($this->configuration['entity_types'] != $this->defaultConfiguration()['entity_types']) {
      throw new \DomainException('The supported entity types for an entity workflow plugin can not be changed.');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityTypes() {
    return array_keys($this->configuration['entity_types']);
  }

  /**
   * {@inheritdoc}
   */
  public function getBundlesForEntityType($entity_type_id) {
    return $this->configuration['entity_types'][$entity_type_id] ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function appliesToEntityTypeAndBundle($entity_type_id, $bundle_id) {
    $bundles_for_entity_type = $this->configuration['entity_types'][$entity_type_id] ?? [];
    return empty($bundles_for_entity_type) || in_array($bundle_id, $bundles_for_entity_type, TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function getDefaultState() {
    return $this->configuration['default_state'];
  }

  /**
   * {@inheritdoc}
   */
  public function getDefaultStateCallback() {
    return $this->configuration['default_state_callback'];
  }

  /**
   * {@inheritdoc}
   */
  public function getDefaultTransition() {
    return $this->configuration['default_transition'];
  }

  /**
   * {@inheritdoc}
   */
  public function getDefaultTransitionLog() {
    return $this->configuration['default_transition_log'];
  }

  /**
   * {@inheritdoc}
   */
  public function getPermissions() {
  }

  /**
   * {@inheritdoc}
   */
  public function getInitialState(ContentEntityInterface $entity = NULL) {
    // Workflows are not tied to entities, but Entity Workflow adds the
    // relationship between Workflows and entities. Entity Workflow needs the
    // entity object to be able to determine the initial state when a default
    // state callback is used.
    if (!($entity instanceof ContentEntityInterface)) {
      throw new \InvalidArgumentException('A content entity object must be supplied.');
    }

    // Check to see if there's a default state callback.
    if (($default_state_callback = $this->getDefaultStateCallback()) && function_exists($default_state_callback)) {
      return $this->getState($default_state_callback($entity));
    }

    // Check to see if there's a default state set.
    if ($default_state = $this->getDefaultState()) {
      return $this->getState($default_state);
    }

    // Fall back to the first state we find in the states list.
    return parent::getInitialState();
  }

  /**
   * {@inheritdoc}
   *
   * @return \Drupal\entity_workflow\EntityWorkflowState
   */
  public function getState($state_id) {
    if (!isset($this->configuration['states'][$state_id])) {
      throw new \InvalidArgumentException("The state '$state_id' does not exist in workflow.");
    }
    return new EntityWorkflowState(
      $this,
      $state_id,
      $this->configuration['states'][$state_id]['label'],
      $this->configuration['states'][$state_id]['weight']
    );
  }

  /**
   * {@inheritdoc}
   *
   * @return \Drupal\entity_workflow\EntityWorkflowTransition
   */
  public function getTransition($transition_id) {
    if (!isset($this->configuration['transitions'][$transition_id])) {
      throw new \InvalidArgumentException("The transition '$transition_id' does not exist in workflow.");
    }
    return new EntityWorkflowTransition(
      $this,
      $transition_id,
      $this->configuration['transitions'][$transition_id]
    );
  }

}
